function [sampleLabels, subspaceBases, returnStatus] = RANSACarrangement(X, subspaceDimensions, varargin)

% Set default parameter values
maxIteration = 10000;
postOptimization = true;
boundaryThreshold = 0.1;
[ambientDimension sampleCount]= size(X);
subspaceCount = length(subspaceDimensions);
consensusThreshold = 1;
angleTolerance = 0;
dataNormalization = true;

% Parse the optional inputs.
if (mod(length(varargin), 2) ~= 0 ),
    error(['Extra Parameters passed to the function ''' mfilename ''' must be passed in pairs.']);
end
parameterCount = length(varargin)/2;

parameterInitialized = false;
for parameterIndex = 1:parameterCount,
    parameterName = varargin{parameterIndex*2 - 1};
    parameterValue = varargin{parameterIndex*2};
    switch lower(parameterName)
        case 'postoptimization'
            postOptimization = parameterValue;
            if postOptimization~=1 && postOptimization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'consensusthreshold'
            consensusThreshold = parameterValue;
            if parameterValue<=0 || parameterValue>=1
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'datanormalization'
            dataNormalization = parameterValue;
            if dataNormalization~=1 && dataNormalization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'maxiteration'
            maxIteration = parameterValue;
            if maxIteration<10
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'boundarythreshold'
            boundaryThreshold = parameterValue;
            if boundaryThreshold<=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        otherwise
            error(['Sorry, the parameter ''' parameterName ''' is not recognized by the function ''' mfilename '''.']);
    end
end


% Normalize data vectors
sampleMagnitudes = zeros(1,sampleCount);
for sampleIndex=1:sampleCount
    sampleMagnitudes(sampleIndex)=norm(X(:,sampleIndex));
end
X = X/max(sampleMagnitudes);
if dataNormalization
    [X, subspaceVariances]=normalize_variance(X);
end

% Compute the number of samples in a minimal sample set
minimalSampleSize = sum(subspaceDimensions);
sampleLabels = [];
subspaceBases = [];
maxConsensus = 0;
for iterationIndex=1:maxIteration
    % Step 1. Generate a minimal sample set
    newSampleIndices = randperm(sampleCount);
    
    % Step 2. Construct bases using the sample set
    samplePointer = 1;
    for subspaceIndex=1:subspaceCount
        samples = X(:,newSampleIndices(samplePointer:samplePointer+subspaceDimensions(subspaceIndex)-1));
        newSubspaceBases{subspaceIndex} = orth(samples);
        samplePointer = samplePointer + subspaceDimensions(subspaceIndex);
    end
    
    % Step 3. Determine the membership of all samples based on the assumed structure
    for subspaceIndex=1:subspaceCount
        distance(subspaceIndex,:) = point_to_space_distance(X,newSubspaceBases{subspaceIndex});
    end
    for sampleIndex=1:sampleCount
        [minDistance, distancePointer]=min(distance(:,sampleIndex));
        if minDistance>boundaryThreshold
            % Outlier
            newSampleLabels(sampleIndex) = -1;
        else
            % Assign subspace label
            newSampleLabels(sampleIndex) = distancePointer;
        end
    end    
    % Test if there are empty groups
    emptyGroupFlag = false;
    for subspaceIndex=1:subspaceCount
        if sum(newSampleLabels==subspaceIndex)==0
            emptyGroupFlag = true;
            break;
        end
    end
    if emptyGroupFlag==true
       %This minimal sample set failed
        continue;
    end     
    
    % Step 4. Recalculate the subspace structures
    currentConsensus = sum(newSampleLabels~=-1)/sampleCount;
    if currentConsensus > maxConsensus
        % Update record
        maxConsensus = currentConsensus;
        sampleLabels = newSampleLabels;
        for subspaceIndex=1:subspaceCount
            samples = X(:,newSampleLabels==subspaceIndex);
            [U,S,V] = svds(samples,subspaceDimensions(subspaceIndex));
            subspaceBases{subspaceIndex} = U;
        end
        if currentConsensus>=consensusThreshold
            % Exceed the threshold, accept this result as final
            return;
        end
    end
end
if isempty(sampleLabels)
    returnStatus = -1;
    return;
end

if postOptimization
    inlierIndex = find(sampleLabels~=-1);
    [sampleLabels(inlierIndex), subspaceBases, returnStatus]=Ksubspaces(X(:,inlierIndex), subspaceDimensions, 'initialBases',subspaceBases);
end

if dataNormalization
    X = normalize_variance(X,subspaceVariances);
    for subspaceIndex=1:subspaceCount
        subspaceX=X(:,find(sampleLabels==subspaceIndex));
        [U,S,V]=svds(subspaceX, subspaceDimensions(subspaceIndex));
        subspaceBases{subspaceIndex} = U;
    end
end